/*
    dltrace 0.5
    Copyright (C) 2005 Richard Johnson <rjohnson@idefense.com>

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation; either version 2 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program; if not, write to the Free Software Foundation, Inc., 59 Temple
    Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "plump.h"
#include "melf.h"
#include "dltrace.h"


int dt_open_process(proc_t *ps)
{
		
	if((plump_open(ps->pid, PLUMP_ACCESS_ALL, PLUMP_MEDIUM_ANY, &ps->plump)) != PLUMP_SUCCESS)
	{
		perror("plump_open");
		return -1;
	}
	if(options.verbose)
		vlog(stderr, "dltrace: process opened [%d] %s\n", ps->pid, ps->argv[0]);

	if(dt_load_binary(ps) < 0)
	{
		vlog(stderr, "error: dt_load_binary %s\n", ps->argv[0]);
		return -1;
	}

	return 0;
}

int dt_library_reloc_syms(proc_t *ps)
{
	uint32_t ptr;
	uint32_t data;
	char *buf;
	
	buf = malloc(HASH_DATASZ);
	if(!buf)
	{
		perror("malloc");
		return -1;
	}

	for(ptr = 0; ptr <= 0xc0000000; ptr += PAGE_SIZE)
	{		
		list_t *l = NULL, *s = NULL;
		uint32_t hash;
		
		if(!(plump_memoryRead(ps->plump, ptr, &data, sizeof(uint32_t), 0) == PLUMP_SUCCESS))
			continue;

		if(!(data == ELFSIG))
			continue;
		
		memset(buf, 0, HASH_DATASZ);
		if(!(buf = dt_mem_get_chunk(ps, ptr, HASH_DATASZ)))
			continue; 

		hash = dt_hash_calc(buf, HASH_DATASZ, HASH_KEY);

		l = ps->liblist;
		while(l->next != NULL)
		{
			library_t *lib = (library_t *)l->data;

			if(!((hash == lib->hash) && (lib->addr == 0)))
			{
				l = l->next;
				continue;
			}
			
			lib->addr = ptr;
			s = lib->symbols;

			if(options.verbose > 1)
				vlog(stderr, "dltrace: library load address [%08x] %s\n", 
					ptr, ((library_t *)l->data)->path);

			while(s->next != NULL)
			{
				symbol_t  *sym = (symbol_t *)s->data;
				uint32_t addr = lib->addr + sym->addr;

				if((sym->addr < lib->addr) && sym->addr != 0)
					sym->addr = addr;
				s = s->next;
			}
			l = l->next;
		}
		free(buf);
			
	}
	
	return 0;
}


/* Memory retrieval functions */
uint8_t *dt_mem_get_chunk(proc_t *ps, uint32_t addr, uint32_t len)
{
	uint8_t *buf;
	
	buf =  malloc(len);
	if(buf < 0)
	{
		perror("malloc");
		return NULL;
	}

	if(!(plump_memoryRead(ps->plump, addr, buf, len, 0) == PLUMP_SUCCESS))
	{
		if(options.verbose > 2)
			vlog(stderr, "warning: could not retrieve memory from 0x%08x\n", addr);	
		free(buf);
		return NULL;

	}

	return buf;
}

uint8_t *dt_mem_get_string(proc_t *ps, uint32_t addr, uint32_t len)
{
	uint8_t *buf = NULL, *str = NULL;
	
	if(len == 0)
		len = 32;
	
	buf = malloc(len + 1);
	if(buf < 0)
	{
		perror("malloc");
		return NULL;
	}

	memset(buf, 0, len + 1);
	if(!(plump_memoryRead(ps->plump, addr, buf, len, 0) == PLUMP_SUCCESS))
	{
		if(options.verbose > 2)
			vlog(stderr, "warning: could not retrieve memory from 0x%08x\n", addr);	
		free(buf);
		return NULL;

	}

	len = strlen(buf);
	for(str = buf; str < buf + len + 1; str++)
	{
		if(*str == 0 || *str > 0x7f || *str < 0x20)
		{
			*str = 0;
			if(str == buf)
				return NULL;
			else
				return buf;
		}		
		if(*str == '\r')
		{
			strncpy(str, "\\r...\0", len - (str - buf));
			return buf;
		}
		else if(*str == '\n')
		{
			if(str == buf + len - 1)
				*str = 0;
			else
				strncpy(str, "\\n...\0", len - (str - buf));
			return buf;
		}
		else if(*str > 0x7f)
		{
			*str = 0;
			return buf;
		}
	}

	return buf;
}

